//----------------------------------------------------------------------------------------------
// BSP : Seeed nRF52 Borads 1.1.1
// Board : Seeed nRF52 Borads / Seeed XIAO nRF52840
//----------------------------------------------------------------------------------------------
// 2023/12/16
// Sleep Current 6.4uA @3.6V
// Note: Remove E5's LDO and connect IN and OUT pins to reduce LDO quiescent current 65uA
//
// nRF52_XIAO Grove-LoRa-E5
//   TX(D6) -----> RX
//   RX(D7) <----- TX
//   3V3    -----> 3V3  Do not supply 5V to LoRa-E5
//   GND    <----> GND

#include <Arduino.h>
#include <Adafruit_TinyUSB.h>
#include <bluefruit.h>          // /Arduino15/packages/Seeeduino/hardware/nrf52/1.1.1/libraries/Bluefruit52Lib/src
#include <Adafruit_SPIFlash.h>  // Need to be deleted /Documents/Arduino/libraries/SdFat

// Built from the P25Q16H datasheet.
// https://gitlab.com/arduino5184213/seeed_xiao_nrf52840/flash_speedtest/-/tree/master
SPIFlash_Device_t const P25Q16H {
  .total_size = (1UL << 21), // 2MiB
  .start_up_time_us = 10000, // Don't know where to find that value  
  .manufacturer_id = 0x85,
  .memory_type = 0x60,
  .capacity = 0x15,
  .max_clock_speed_mhz = 55,
  .quad_enable_bit_mask = 0x02, // Datasheet p. 27
  .has_sector_protection = 1,   // Datasheet p. 27
  .supports_fast_read = 1,      // Datasheet p. 29
  .supports_qspi = 1,           // Obviously
  .supports_qspi_writes = 1,    // Datasheet p. 41
  .write_status_register_split = 1, // Datasheet p. 28
  .single_status_byte = 0,      // 2 bytes
  .is_fram = 0,                 // Flash Memory
};
Adafruit_FlashTransport_QSPI flashTransport;
Adafruit_SPIFlash flash(&flashTransport);

#define HICHG         22        // Charge current Hi:50mA Lo:100mA        D22 (13u)
#define CHG           23        // Charge LED Lo:ON                       D23 (17u)
#define VBAT_ENABLE   14        // Battery voltage read enable  Lo:enable D14 (14u)   
#define VBAT_READ     32        // Battery voltage read pin               D32 (31u)
#define TIMEOUT    15000        // Response timeout ms
#define LP_TIME    10000        // Period of Low Power ms

char devaddr[9] = {"510068D4"};
char testrfcfg[64] = {"AT+TEST=RFCFG,922.1,SF7,250,8,8,13,ON,OFF,OFF\r\n"};   // 13dBm=20mW, SF7=11000bps
//char testrfcfg[64] = {"AT+TEST=RFCFG,922.1,SF7,250,8,8,0,ON,OFF,OFF\r\n"};   // 0dBm=1mW, SF7=11000bps
//char testrfcfg[64] = {"AT+TEST=RFCFG,922,SF12,125,8,8,13,ON,OFF,OFF\r\n"};   // SF12=250bps
char recv_buff[512];
bool is_exist = true;

void setup(void)
{
  // Enable DC-DC converter Mode
  NRF_POWER->DCDCEN = 1;            // Enable DC/DC converter for REG1 stage

  // on board Flash enter to Deep Power-Down Mode
  flashTransport.begin();
  flashTransport.runCommand(0xB9);  // enter deep power-down mode
  delayMicroseconds(5);             // tDP=3uS
  flashTransport.end();

  // SerialSerial port initialization
  Serial.begin(115200);
  while (!Serial);
  delay(1000);
  Serial.println("communication test TX");

  // Pin initialization
//  pinMode(VBAT_READ, INPUT); 
//  pinMode(HICHG, OUTPUT);
  pinMode(CHG, INPUT_PULLUP);       // sleep current increases significantly whithout pull-up
  pinMode(VBAT_ENABLE, OUTPUT);   
  pinMode(LED_RED, OUTPUT);
  pinMode(LED_GREEN, OUTPUT);
  pinMode(LED_BLUE, OUTPUT); 

//  digitalWrite(HICHG, LOW);         // Charge current 100mA
  digitalWrite(VBAT_ENABLE, HIGH);   // Battery voltage read disable   
  digitalWrite(LED_RED, HIGH);
  digitalWrite(LED_GREEN, HIGH);
  digitalWrite(LED_BLUE, HIGH);

  // ADC initialization　See wireing_analog_nRF52.c:73
  analogReference(AR_DEFAULT);      // Default 0.6V*6=3.6V
  analogReadResolution(12);         // wireing_analog_nRF52.c:39
  analogSampleTime(10);             // default 3uS　3,5,10,15,20,40
  
  // UART initialization
  Serial1.begin(115200);
  delay(1000);

  // LoRa initialization
  Serial.print("send check response!\r\n");
  at_command("AT\r\n");
  if(!at_response(1, TIMEOUT)) {
    Serial.println("AT timeout");
    is_exist = false;
  }
  at_command("AT+ID\r\n");
  if(!at_response(3, TIMEOUT)) {
    Serial.println("AT+ID timeout");
    is_exist = false;
  }
  at_command("AT+MODE=TEST\r\n");
  if(!at_response(1, TIMEOUT)) {
    Serial.println("AT+MODE=TEST timeout");
    is_exist = false;    
  }
  at_command(testrfcfg);  
  if(!at_response(1, TIMEOUT)) { 
    Serial.println("AT+TEST=RFCFG timeout");
    is_exist = false;    
  }
  if(is_exist == false) {
    Serial.println("E5 response error");
  } 
}

void loop(void)
{
  if (is_exist)
  {
    // Battery voltage read
    digitalWrite(VBAT_ENABLE, LOW);      // Battery voltage read enable   
    int Vadc = 0;                        // Battery voltage averaging
    for (int i = 0; i < 16; i++) 
    {
      Vadc = Vadc + analogRead(VBAT_READ);
    }
    digitalWrite(VBAT_ENABLE, HIGH);     // Battery voltage read disable 
    
    Vadc = Vadc / 16;
    float Vbatt = 2.961 * 3.6 * Vadc / 4096;   // Vref=3.6  ratio=2.961

    // Ttansmit DATA    
    uint32_t ms = millis();
    Serial.println(ms);           
    char data1[32];
    char data2[32];
    char cmd[128];
    memset(data1, 0, sizeof(data1));
    memset(data2, 0, sizeof(data2));
    sprintf(data1, "%08d", ms / 1000);
    sprintf(data2, "%04d", int(Vbatt*1000));       
    sprintf(cmd, "AT+TEST=TXLRPKT,\"%s%s%s\"\r\n", devaddr, data1, data2);      
    at_command(cmd);
    if(!at_response(2, TIMEOUT)) {
      Serial.println("AT+TEST=TXLRPKT timeout");    
    }

    // put E5 into sleep mode
    at_command("AT+TEST=STOP\r\n");                           // TEST MODE stop
    if(!at_response(1, TIMEOUT)) {
      Serial.println("AT+TEST=STOP timeout");    
    }
    at_command("AT+LOWPOWER\r\n");                            // E5 Low Power
    if(!at_response(1, TIMEOUT)) {
      Serial.println("AT+LOWPOWER SLEEP timeout");
    }    
    Serial1.end();                                            // UART disable
    delay(LP_TIME);                                           // nRF52 Light Sleep, wake up

    // Wake up
    Serial1.begin(115200);                                    // UART enable
    delay(100);       
    at_command("A");                                          // E5 wake up
    if(!at_response(1, TIMEOUT)) {
      Serial.println("AT+LOWPOWER WAKEPU timeout");
    }
    delay(5); 
    at_command("AT+MODE=TEST\r\n");                           // enter TEST mode
    if(!at_response(1, TIMEOUT)) {
      Serial.println("AT+MODE=TEST timeout");   
    }
    at_command(testrfcfg);
    if(!at_response(1, TIMEOUT)) { 
      Serial.println("AT+TEST=RFCFG timeout");   
    }                  
  }
}

void at_command(char *p_cmd)
{ 
  // send command
  digitalWrite(LED_RED, LOW);
  delay(10);
  Serial1.print(p_cmd);
  digitalWrite(LED_RED, HIGH);
}

int at_response(int line_numbers, uint32_t timeout)
{
  // receive response
  digitalWrite(LED_GREEN, LOW);    
  int startMillis = millis();
  int linefeed = 0;
  char c;
  int index = 0; 
    do
    {
      while(Serial1.available() > 0)
      {
        c = Serial1.read();
        recv_buff[index++] = c;
        Serial.print(c);
        if(c == '\n') {
          linefeed++;
          if(linefeed == line_numbers) {    // check response line numbers
            digitalWrite(LED_GREEN, HIGH);
            return 1;
          }
        }
      } 
    } while(linefeed < line_numbers && millis() - startMillis < timeout);  // check timeout       
  digitalWrite(LED_GREEN, HIGH);
  return 0;
}
